import hashlib
import shutil
import subprocess
from pathlib import Path

import gitignorant as gi


def enumerate_files(root: Path, rules: list[gi.Rule]):
    """Enumerate all files under `root` according to (inverse) gitignore rules."""

    assert root.is_dir()
    dirs = [root]
    files = []
    i = 0
    while i < len(dirs):
        dir = dirs[i]
        for name in dir.iterdir():
            f = dir / name
            relpath = f.relative_to(root)
            match = gi._find_match(rules, str(relpath), is_dir=True)
            if f.is_dir():
                if match is not False:
                    dirs.append(f)
            elif match is True:
                files.append(f)
        i += 1
    return files


def hash_files(files: list[Path]) -> str:
    """Compute a single hash for a bunch of files."""

    cmd = ["sha256sum"] + (sorted(files))
    proc = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
    if proc.returncode != 0:
        raise RuntimeError("Could not hash files")
    m = hashlib.blake2b(digest_size=12)
    m.update(proc.stdout)
    return m.hexdigest()


def copy_files(from_: Path, to: Path, files: list[Path]):
    """Copy `files` and their parent directories from `from_` to `to`."""

    dirs = set()
    for f in files:
        dirs.update(f.parents)
    ignores = set(files) | dirs

    def ignore(path, contents):
        p = Path(path)
        return [f for f in contents if p / f not in ignores]

    shutil.copytree(from_, to, ignore=ignore)



import logging
import os
from pathlib import Path

import click
import gitignorant as gi

from .config import Config, find_pyproject_toml
from .files import copy_files, enumerate_files, hash_files

# Since this module controls the application, we get the root logger here instead of the
# one specified by __name__
log = logging.getLogger("brezn")


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option("-v", "--verbose", count=True)
@click.option("-c", "--config")
@click.pass_context
def main(ctx, verbose, config):
    """Something layered on top of hydra."""

    if verbose >= 1:
        logging.basicConfig()
        log.setLevel(logging.INFO)
        if verbose >= 2:
            log.setLevel(logging.DEBUG)

    pyproject_toml = find_pyproject_toml()
    if pyproject_toml is None:
        raise click.ClickException("Could not find a pyproject.toml")

    log.info(f"Load configuration from {pyproject_toml}")
    ctx.obj = Config.from_pyproject_toml(pyproject_toml)
    log.info(f"Configuration {ctx.obj}")


@main.command(context_settings={"ignore_unknown_options": True})
@click.argument("file", type=click.Path(exists=True, executable=True, dir_okay=False))
@click.argument("options", nargs=-1)
@click.pass_context
def run(ctx, file, options):
    """Run the provided command after copying the environment."""

    config: Config = ctx.obj
    file = Path(file)

    project_root = config.project_root
    env_dir = config.env_dir
    if not env_dir.is_absolute():
        env_dir = (project_root / config.env_dir).resolve()

    # Parse the rules for which files to copy
    rules = [rule for r in config.files if (rule := gi.try_parse_rule(r)) is not None]
    # Ignore anything in the environment directory
    if env_dir.is_relative_to(project_root):
        rules += [
            gi.Rule(negative=True, content="/" + str(env_dir.relative_to(project_root)))
        ]
    files = enumerate_files(project_root, rules)

    # Copy the project files into the environment directory
    env_dir.mkdir(exist_ok=True, parents=True)
    files_hash = hash_files(files)
    copy_root = env_dir / files_hash
    if not copy_root.is_dir():
        copy_files(project_root, copy_root, files)

    # Replace the current process to forward stdout, stderr, exit code and anything else
    # to and from the program as best as possible.
    if file.is_relative_to(project_root):
        file = file.relative_to(project_root)
    os.execvp((copy_root / file).absolute(), [file] + list(options))



import tomllib
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Config:
    @staticmethod
    def from_pyproject_toml(file: Path):
        with file.open("rb") as f:
            pyproject = tomllib.load(f)
        brezn_table = pyproject.get("tool", {}).get("brezn", {})

        return Config(
            project_root=file.parent,
            env_dir=Path(brezn_table.get("env_dir", ".brezn")),
            files=brezn_table.get("files", []),
        )

    project_root: Path
    env_dir: Path
    files: list[str]


def find_pyproject_toml():
    dir = Path.cwd().resolve()

    for root in [dir] + list(dir.parents):
        pp_toml = root / "pyproject.toml"
        if pp_toml.is_file():
            return pp_toml
